#include <vector>
#include <time.h>

#include<vcg/simplex/vertexplus/base.h>
#include<vcg/simplex/faceplus/base.h>
#include<vcg/simplex/face/topology.h>
#include<vcg/complex/trimesh/base.h>
#include<wrap/io_trimesh/import_ply.h>

// topology computation
#include<vcg/complex/trimesh/update/topology.h>
#include<vcg/complex/trimesh/update/bounding.h>
#include<vcg/complex/trimesh/update/flag.h>
#include<vcg/complex/trimesh/clean.h>

// half edge iterators
#include<vcg/simplex/face/pos.h>


using namespace vcg;

class MyEdge;    // dummy prototype never used
class MyFace;
class MyVertex;

class MyVertex  : public VertexSimp2< MyVertex, MyEdge, MyFace, vert::Coord3f, vert::BitFlags  >{};
class MyFace    : public FaceSimp2  < MyVertex, MyEdge, MyFace, face::VertexRef, face::FFAdj, face::Mark, face::BitFlags > {};

class MyMesh : public tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace > >{};


int main(int argc,char **argv ){

	MyMesh m;
  int t0=clock();

	if(tri::io::ImporterPLY<MyMesh>::Open(m,argv[1])!=0)
		{
      printf("Error reading file  %s\n",argv[1]);
			exit(0);
		}

    int t4=clock();

  printf("Mesh has %i vert and %i faces\n", m.vn,m.fn);
  printf("Mesh is %f x %f x %f \n", m.bbox.max[0]-m.bbox.min[0],
                                    m.bbox.max[1]-m.bbox.min[1],
                                    m.bbox.max[2]-m.bbox.min[2]);
  tri::UpdateBounding<MyMesh>::Box(m);
  printf("Mesh is %f x %f x %f \n", m.bbox.DimX(),m.bbox.DimY(),m.bbox.DimZ());

  tri::UpdateTopology<MyMesh>::FaceFace(m);
  int t1=clock();
  printf("%i msec to load and update the mesh\n",t1-t0);
/*
This sample show two methods to count the non two-manifold edges in a loaded mesh 
and report their 'complexity', i.e. the number of faces incident on the same edge.
When computing the face-face topology the library connects in a single chain 
all the faces incident on a same edge, managing non manifold edges in a consistent way.

The chain of faces is not spatially ordered (to reflect the fact that the topological 
relations should be independent from the geometric realization of the mesh.

In the first approach we scan all the faces and for each edge we follow the chain of 
faces incident on them trhough the FFp and FFi operators. When this face chain is longer 
than 2 elements and the visit is started form the face with the higher face pointer, 
then the non manifold edge is reported. 
Note that the edge chain is traversed many times (once for every face incident on it).
*/

  MyMesh::FaceIterator fi;
  for(fi=m.face.begin();fi!=m.face.end();++fi)               
    for(int i=0;i<3;++i)
    {
      bool isMax=true;
      MyMesh::FacePointer adj=(*fi).FFp(i);
      int adjInd=(*fi).FFi(i);        
      int complexity=1;
      do {
        if( adj > &*fi ) isMax=false;
        MyMesh::FacePointer adjNext=adj->FFp(adjInd);
        adjInd=adj->FFi(adjInd);        
        ++complexity;
        adj=adjNext;
      } while(adj != &*fi);

      if(isMax && complexity>2) 
        printf("Found a non manifold edge (%i-%i) of %i\n",
        (*fi).V0(i)-&*m.vert.begin() ,(*fi).V1(i)-&*m.vert.begin(), complexity );
    }
  int t2=clock();
  printf("%i msec to find non 2 manifold edges\n",t2-t1);
  
/*
In the second approach we explicitly mark the face chains that we have already visited 
(and eventually reported).
Note that the basic vertex and face classes do not allow per-edge storage so we have 
to use three user defined bits for each face. These bits are 'allocated' from the  
pool of available bits (and then released). In this way each face-on-an-edge loop 
is traversed only once.
*/
    printf("\n------------------------------\n\n");

    
    int edgeBit[3];
    edgeBit[0]=MyFace::NewBitFlag();
    edgeBit[1]=MyFace::NewBitFlag();
    edgeBit[2]=MyFace::NewBitFlag();

    for(fi=m.face.begin();fi!=m.face.end();++fi)
      for(int i=0;i<3;++i)
        (*fi).ClearUserBit(edgeBit[i]);
    
    for(fi=m.face.begin();fi!=m.face.end();++fi)
      for(int i=0;i<3;++i)
      {
        if(!(*fi).IsUserBit(edgeBit[i]))
        {
          (*fi).SetUserBit(edgeBit[i]);
          MyMesh::FacePointer adj=(*fi).FFp(i);
          int adjInd=(*fi).FFi(i);        
          int complexity=1;
          do {
            adj->SetUserBit(edgeBit[adjInd]);
            MyMesh::FacePointer adjNext=adj->FFp(adjInd);
            adjInd=adj->FFi(adjInd);        
            adj=adjNext;
            ++complexity;
           } while(adj != &*fi);
        if(complexity>2)
            printf("Found a non manifold edge (%i-%i) of %i\n",
              (*fi).V0(i)-&*m.vert.begin() ,(*fi).V1(i)-&*m.vert.begin(), complexity );    
        }
      }
    
    MyFace::DeleteBitFlag(edgeBit[2]);
    MyFace::DeleteBitFlag(edgeBit[1]);
    MyFace::DeleteBitFlag(edgeBit[0]);
    
    int t3=clock();
    printf("%i msec to find non 2 manifold edges\n",t3-t2);
    
  return 0;
}







